__all__ = [
    "Diagnostic",
    "DiagnosticCollection",
    "DiagnosticError",
    "DiagnosticErrorSummary",
]


from dataclasses import dataclass, field, replace
from textwrap import indent
from types import TracebackType
from typing import Any, Iterator, List, Literal, Optional, Type

from beet import BubbleException, TextFileBase
from tokenstream import UNKNOWN_LOCATION, SourceLocation

from .error import MechaError
from .utils import underline_code


@dataclass
class Diagnostic(MechaError):
    """Exception that can be raised to report messages."""

    level: Literal["info", "warn", "error"]
    message: str
    notes: List[str] = field(default_factory=list)

    rule: Optional[str] = None
    hint: Optional[str] = None
    filename: Optional[str] = None
    file: Optional[TextFileBase[Any]] = None
    location: SourceLocation = UNKNOWN_LOCATION
    end_location: SourceLocation = UNKNOWN_LOCATION

    def __str__(self) -> str:
        return self.format_message()

    def format_message(self) -> str:
        """Return the formatted message."""
        message = self.message
        if self.rule:
            message += f" ({self.rule})"
        return message

    def format_notes(self) -> str:
        """Return the formatted notes."""
        return "Notes:\n" + "\n".join(
            "  - " + indent(note, "    ")[4:] for note in self.notes
        )

    def format_location(self) -> str:
        """Return the formatted location of the reported message."""
        if self.filename:
            location = self.filename
            if not self.location.unknown:
                location += f":{self.location.lineno}:{self.location.colno}"
        elif not self.location.unknown:
            location = f"line {self.location.lineno}, column {self.location.colno}"
            if self.hint:
                location = f'File "{self.hint}", {location}'
        elif self.hint:
            location = self.hint
        else:
            location = ""
        return location

    def format_code(self, code: str) -> Optional[str]:
        """Return the formatted code."""
        if self.location.unknown:
            return None
        return underline_code(
            code,
            self.location,
            self.location if self.end_location.unknown else self.end_location,
        )

    def with_defaults(
        self,
        rule: Optional[str] = None,
        hint: Optional[str] = None,
        filename: Optional[str] = None,
        file: Optional[TextFileBase[Any]] = None,
        location: SourceLocation = UNKNOWN_LOCATION,
        end_location: SourceLocation = UNKNOWN_LOCATION,
    ) -> "Diagnostic":
        """Set default values for unspecified attributes."""
        if not self.location.unknown:
            location = self.location
            end_location = self.end_location
        return replace(
            self,
            rule=self.rule or rule,
            hint=self.hint or hint,
            filename=self.filename or filename,
            file=self.file or file,
            location=location,
            end_location=end_location,
        )


@dataclass
class DiagnosticCollection(MechaError):
    """Exception that can be raised to group multiple diagnostics."""

    exceptions: List[Diagnostic] = field(default_factory=list)

    rule: Optional[str] = None
    hint: Optional[str] = None
    filename: Optional[str] = None
    file: Optional[TextFileBase[Any]] = None

    def add(self, exc: Diagnostic) -> Diagnostic:
        """Add diagnostic."""
        exc = exc.with_defaults(
            rule=self.rule,
            hint=self.hint,
            filename=self.filename,
            file=self.file,
        )
        self.exceptions.append(exc)
        return exc

    def extend(self, other: "DiagnosticCollection"):
        """Combine diagnostics from another collection."""
        self.exceptions.extend(other.exceptions)

    def clear(self):
        """Clear all the diagnostics."""
        self.exceptions.clear()

    @property
    def error(self) -> bool:
        """Return true if the diagnostics contain at least one error."""
        for exc in self.exceptions:
            if exc.level == "error":
                return True
        return False

    def get_all_errors(self) -> Iterator[Diagnostic]:
        """Yield all the diagnostics with a severity level of "error"."""
        for exc in self.exceptions:
            if exc.level == "error":
                yield exc

    def __enter__(self) -> "DiagnosticCollection":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        if not exc_type:
            if self.exceptions:
                raise self

    def __str__(self) -> str:
        term = (
            "error"
            if all(exc.level == "error" for exc in self.exceptions)
            else "diagnostic"
        )
        term += "s" * (len(self.exceptions) > 1)
        return f"Reported {len(self.exceptions)} {term}."


class DiagnosticError(MechaError):
    """Raised with a collection of error diagnostics."""

    diagnostics: DiagnosticCollection

    def __init__(self, diagnostics: DiagnosticCollection):
        super().__init__(diagnostics)
        self.diagnostics = diagnostics

    def __str__(self) -> str:
        details = "\n".join(
            f"{diagnostic.format_location()}: {diagnostic.format_message()}"
            for diagnostic in self.diagnostics.exceptions
        )
        return f"{self.diagnostics}\n\n{details}"


class DiagnosticErrorSummary(BubbleException):
    """Raised for showing a summary of how many errors occurred."""

    diagnostics: DiagnosticCollection

    def __init__(self, diagnostics: DiagnosticCollection):
        super().__init__(diagnostics)
        self.diagnostics = diagnostics

    def __str__(self) -> str:
        return str(self.diagnostics)



__all__ = [
    "wrap_backslash_continuation",
    "BACKSLASH_CONTINUATION_REGEX",
]


import re
from itertools import islice

from tokenstream import INITIAL_LOCATION, SourceLocation

BACKSLASH_CONTINUATION_REGEX = re.compile(r"(\\[ \t]*\r?\n[ \t]*)")


def wrap_backslash_continuation(
    source: str,
) -> tuple[str, list[SourceLocation], list[SourceLocation]]:
    it = iter(BACKSLASH_CONTINUATION_REGEX.split(source))
    text = next(it)

    result = [text]
    source_mappings: list[SourceLocation] = []
    preprocessed_mappings: list[SourceLocation] = []

    source_location = INITIAL_LOCATION.skip_over(text)
    preprocessed_location = source_location

    while True:
        try:
            backslash, text = islice(it, 2)
        except ValueError:
            break

        source_location = source_location.skip_over(backslash)
        source_mappings.append(source_location)
        preprocessed_mappings.append(preprocessed_location)

        result.append(text)
        source_location = source_location.skip_over(text)
        preprocessed_location = preprocessed_location.skip_over(text)

    return "".join(result), source_mappings, preprocessed_mappings



__all__ = [
    "QuoteHelper",
    "QuoteHelperWithUnicode",
    "JsonQuoteHelper",
    "NbtQuoteHelper",
    "InvalidEscapeSequence",
    "normalize_whitespace",
    "string_to_number",
    "number_to_string",
    "underline_code",
]


import re
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from tokenstream import InvalidSyntax, SourceLocation, Token, set_location

from .error import MechaError

ESCAPE_REGEX = re.compile(r"\\.")
UNICODE_ESCAPE_REGEX = re.compile(r"\\(?:u([0-9a-fA-F]{4})|.)")
AVOID_QUOTES_REGEX = re.compile(r"^[0-9A-Za-z_\.\+\-]+$")
QUOTE_REGEX = re.compile(r"\"|'")

WHITESPACE_REGEX = re.compile(r"\s+")

NOTABS_REGEX = re.compile(r"[^\t]")


def normalize_whitespace(op: str) -> str:
    """Replace whitespace with underscores."""
    return WHITESPACE_REGEX.sub("_", op).strip("_")


def string_to_number(string: str) -> Union[int, float]:
    """Helper for converting numbers to string and keeping their original type."""
    return float(string) if "." in string or "e" in string else int(string)


def number_to_string(number: Union[int, float]) -> str:
    """Helper for converting numbers to string and removing scientific notation."""
    value = str(number)
    if (index := value.find("e")) != -1:
        exponent = int(value[index + 1 :])
        value = f"{number:.{max(index - 1 - (value[1] == '.') - exponent, 1)}f}"
    return value


class InvalidEscapeSequence(MechaError):
    """Raised when a QuotedStringHandler encounters an invalid escape sequence."""

    characters: str
    index: int

    def __init__(self, characters: str, index: int):
        super().__init__(characters, index)
        self.characters = characters
        self.index = index


@dataclass
class QuoteHelper:
    """Helper for removing quotes and interpreting escape sequences."""

    escape_regex: "re.Pattern[str]" = ESCAPE_REGEX
    avoid_quotes_regex: "re.Pattern[str]" = AVOID_QUOTES_REGEX

    escape_sequences: Dict[str, str] = field(default_factory=dict)
    unquote_only_escape_characters: List[str] = field(default_factory=list)
    escape_characters: Dict[str, str] = field(init=False)

    def __post_init__(self):
        self.escape_characters = {
            v: k
            for k, v in self.escape_sequences.items()
            if k not in self.unquote_only_escape_characters
        }

    def unquote_string(self, token: Token) -> str:
        """Remove quotes and substitute escaped characters."""
        if not token.value.startswith(('"', "'")):
            return token.value
        try:
            return self.escape_regex.sub(
                lambda match: self.handle_substitution(token, match),
                token.value[1:-1],
            )
        except InvalidEscapeSequence as exc:
            location = token.location.with_horizontal_offset(exc.index + 1)
            end_location = location.with_horizontal_offset(len(exc.characters))
            exc = InvalidSyntax(f"Invalid escape sequence {exc.characters}.")
            raise set_location(exc, location, end_location)

    def handle_substitution(self, token: Token, match: "re.Match[str]") -> str:
        """Handle escaped character sequence."""
        characters = match[0]

        if characters == "\\" + token.value[0]:
            return token.value[0]

        try:
            return self.escape_sequences[characters]
        except KeyError:
            raise InvalidEscapeSequence(characters, match.start()) from None

    def quote_string(self, value: str, quote: str = '"') -> str:
        """Wrap the string in quotes if it can't be represented unquoted."""
        if self.avoid_quotes_regex.match(value):
            return value
        value = self.handle_quoting(value)
        return quote + value.replace(quote, "\\" + quote) + quote

    def handle_quoting(self, value: str) -> str:
        """Handle escape characters during quoting."""
        for match, seq in self.escape_characters.items():
            value = value.replace(match, seq)
        return value


@dataclass
class QuoteHelperWithUnicode(QuoteHelper):
    """Quote helper that handles unicode escape sequences."""

    escape_regex: "re.Pattern[str]" = UNICODE_ESCAPE_REGEX

    def handle_substitution(self, token: Token, match: "re.Match[str]") -> str:
        if unicode_hex := match[1]:
            return chr(int(unicode_hex, 16))
        return super().handle_substitution(token, match)

    def handle_quoting(self, value: str) -> str:
        value = super().handle_quoting(value)

        def escape_char(char: str) -> str:
            codepoint = ord(char)
            if codepoint < 128:
                return char
            return f"\\u{codepoint:04x}"

        return "".join(escape_char(c) for c in value)


@dataclass
class JsonQuoteHelper(QuoteHelperWithUnicode):
    """Quote helper used for json."""

    escape_sequences: Dict[str, str] = field(
        default_factory=lambda: {
            r"\\": "\\",
            r"\f": "\f",
            r"\n": "\n",
            r"\r": "\r",
            r"\t": "\t",
        }
    )


@dataclass
class NbtQuoteHelper(QuoteHelperWithUnicode):
    """Quote helper used for snbt."""

    escape_sequences: Dict[str, str] = field(
        default_factory=lambda: {
            r"\\": "\\",
            r"\b": "\b",
            r"\f": "\f",
            r"\n": "\n",
            r"\r": "\r",
            r"\s": " ",
            r"\t": "\t",
        }
    )
    unquote_only_escape_characters: List[str] = field(default_factory=lambda: [r"\s"])

    def quote_string(self, value: str, quote: Optional[str] = None) -> str:
        if not quote:
            found = QUOTE_REGEX.search(value)
            quote = "'" if found and found.group() == '"' else '"'
        value = super().handle_quoting(value)
        return quote + value.replace(quote, "\\" + quote) + quote


def underline_code(
    source: str,
    location: SourceLocation,
    end_location: SourceLocation,
    padding: int = 1,
) -> str:
    """Underline code."""
    pos, lineno, colno = location
    end_pos, end_lineno, end_colno = end_location

    view_begin = pos
    view_end = end_pos

    for _ in range(padding + 1):
        try:
            view_begin = source.rindex("\n", 0, max(view_begin - 1, 0)) + 1
        except ValueError:
            view_begin = 0
        try:
            view_end = source.index("\n", min(view_end + 1, len(source)))
        except ValueError:
            view_end = len(source)

    view = source[view_begin:view_end].splitlines()
    view_start_line = lineno - source[view_begin:pos].count("\n")
    gutter = [f"{i + view_start_line} |" for i in range(len(view))]

    for line in reversed(range(lineno, end_lineno + 1)):
        index = line - view_start_line
        code = view[index] if index < len(view) else ""
        start = colno if line == lineno else 1
        stop = end_colno if line == end_lineno else len(code) + 1

        if start >= stop:
            stop = start + 1

        start = max(start, len(code) - len(code.lstrip()) + 1)
        stop = min(stop, len(code.rstrip()) + 1)

        if start < stop:
            underline = NOTABS_REGEX.sub(" ", code[: start - 1]) + "^" * (stop - start)
            view.insert(index + 1, underline)
            gutter.insert(index + 1, ":")

    gutter_size = max(max(len(g) for g in gutter), 8)
    return "\n".join(f"{g.rjust(gutter_size)}  {text}" for g, text in zip(gutter, view))



"""Plugin for handling nested commands."""

__all__ = [
    "NestedCommandsTransformer",
    "NestingOptions",
    "nesting",
    "parse_nested_root",
]


from dataclasses import dataclass, field, replace
from importlib.resources import files
from typing import Generator, List, cast

from beet import Context, Function
from beet import Generator as BeetGenerator
from beet import configurable
from beet.core.utils import required_field
from pydantic import BaseModel
from tokenstream import InvalidSyntax, TokenStream, set_location

from mecha import (
    AstChildren,
    AstCommand,
    AstResourceLocation,
    AstRoot,
    CommandTree,
    CompilationDatabase,
    Diagnostic,
    Mecha,
    MutatingReducer,
    consume_line_continuation,
    delegate,
    rule,
)
from mecha.contrib.nested_location import NestedLocationResolver


class NestingOptions(BaseModel):
    generate_execute: str = "nested_execute_{incr}"
    generate_macro: str = "nested_macro_{incr}"
    generate_return: str = "nested_return_{incr}"


def beet_default(ctx: Context):
    ctx.require(nesting)


@configurable(validator=NestingOptions)
def nesting(ctx: Context, opts: NestingOptions):
    mc = ctx.inject(Mecha)

    mc.spec.multiline = True

    commands_json = files("mecha.resources").joinpath("nesting.json").read_text()
    mc.spec.add_commands(CommandTree.model_validate_json(commands_json))

    mc.spec.parsers["nested_root"] = parse_nested_root
    mc.spec.parsers["command:argument:mecha:nested_root"] = delegate("nested_root")

    mc.transform.extend(
        NestedCommandsTransformer(
            generate=ctx.generate,
            database=mc.database,
            generate_execute_template=opts.generate_execute,
            generate_macro_template=opts.generate_macro,
            generate_return_template=opts.generate_return,
            nested_location_resolver=ctx.inject(NestedLocationResolver),
        )
    )


def parse_nested_root(stream: TokenStream) -> AstRoot:
    """Parse nested root."""
    with stream.syntax(colon=r":"):
        colon = stream.expect("colon")

    if not consume_line_continuation(stream):
        exc = InvalidSyntax("Expected non-empty block.")
        raise set_location(exc, colon)

    level, command_level = stream.indentation[-2:]

    commands: List[AstCommand] = []

    with (
        stream.intercept("newline"),
        stream.provide(
            scope=(),
            line_indentation=command_level,
        ),
    ):
        while True:
            commands.append(delegate("root_item", stream))

            # The command parser consumes the trailing newline so we need to rewind
            # to be able to use "consume_line_continuation()".
            while (token := stream.peek()) and not token.match("newline", "eof"):
                stream.index -= 1

            with stream.provide(multiline=True, line_indentation=level):
                if not consume_line_continuation(stream):
                    break

    node = AstRoot(commands=AstChildren(commands))
    return set_location(node, commands[0], commands[-1])


@dataclass
class NestedCommandsTransformer(MutatingReducer):
    """Transformer that handles nested commands."""

    generate: BeetGenerator = required_field()
    database: CompilationDatabase = required_field()
    generate_execute_template: str = required_field()
    generate_macro_template: str = required_field()
    generate_return_template: str = required_field()
    nested_location_resolver: NestedLocationResolver = required_field()

    identifier_map: dict[str, str] = field(
        default_factory=lambda: {
            "function:name:commands": "function:name",
            "function:name:arguments:commands": "function:name:arguments",
            "function:name:with:block:sourcePos:commands": "function:name:with:block:sourcePos",
            "function:name:with:block:sourcePos:path:commands": "function:name:with:block:sourcePos:path",
            "function:name:with:entity:source:commands": "function:name:with:entity:source",
            "function:name:with:entity:source:path:commands": "function:name:with:entity:source:path",
            "function:name:with:storage:source:commands": "function:name:with:storage:source",
            "function:name:with:storage:source:path:commands": "function:name:with:storage:source:path",
            "append:function:name:commands": "function:name",
            "prepend:function:name:commands": "function:name",
        }
    )

    def emit_function(self, path: str, root: AstRoot):
        """Helper method for emitting nested commands into a separate function."""
        function = Function(original=self.database.current.original)
        self.generate(path, function)
        self.database[function] = replace(
            self.database[self.database.current],
            ast=root,
            resource_location=path,
        )
        self.database.enqueue(function, self.database.step + 1)

    @rule(AstCommand, identifier="execute:subcommand")
    def nesting_execute_run(self, node: AstCommand):
        if node.arguments and isinstance(command := node.arguments[0], AstCommand):
            if command.identifier == "execute:run:subcommand":
                return command.arguments[0]
        return node

    @rule(AstCommand, identifier="execute:run:subcommand")
    @rule(AstCommand, identifier="return:run:subcommand")
    def nesting_run_function(self, node: AstCommand):
        if isinstance(command := node.arguments[0], AstCommand):
            if command.identifier in self.identifier_map:
                yield from self.handle_function(command)
                command = replace(
                    command,
                    identifier=self.identifier_map[command.identifier],
                    arguments=AstChildren(command.arguments[:-1]),
                )
                return replace(node, arguments=AstChildren([command]))

        return node

    @rule(AstCommand, identifier="execute:commands")
    @rule(AstCommand, identifier="return:commands")
    def nesting_execute_commands(self, node: AstCommand):
        root = cast(AstRoot, node.arguments[0])

        if node.identifier == "execute:commands":
            generate_template = self.generate_execute_template
            identifier = "execute:run:subcommand"
        else:
            generate_template = self.generate_return_template
            identifier = "return:run:subcommand"

        single_command = None
        for command in root.commands:
            if command.compile_hints.get("skip_execute_inline_single_command"):
                continue
            if single_command is None:
                single_command = command
            else:
                single_command = None
                break

        if single_command:
            subcommand = single_command

            if (
                node.identifier == "execute:commands"
                and subcommand.identifier == "execute:subcommand"
            ):
                return subcommand.arguments[0]

        else:
            namespace, resolved = self.nested_location_resolver.resolve()
            path = self.generate.format(f"{namespace}:{resolved}/{generate_template}")

            self.emit_function(path, root)

            subcommand = AstCommand(
                identifier="function:name",
                arguments=AstChildren([AstResourceLocation.from_value(path)]),
            )

        return AstCommand(
            identifier=identifier,
            arguments=AstChildren([subcommand]),
        )

    @rule(AstCommand, identifier="with:block:sourcePos:commands")
    @rule(AstCommand, identifier="with:block:sourcePos")
    @rule(AstCommand, identifier="with:block:sourcePos:path:commands")
    @rule(AstCommand, identifier="with:block:sourcePos:path")
    @rule(AstCommand, identifier="with:entity:source:commands")
    @rule(AstCommand, identifier="with:entity:source")
    @rule(AstCommand, identifier="with:entity:source:path:commands")
    @rule(AstCommand, identifier="with:entity:source:path")
    @rule(AstCommand, identifier="with:storage:source:commands")
    @rule(AstCommand, identifier="with:storage:source")
    @rule(AstCommand, identifier="with:storage:source:path:commands")
    @rule(AstCommand, identifier="with:storage:source:path")
    def nesting_macro_commands(self, node: AstCommand):
        identifier = node.identifier.split(":")
        if not identifier[-1] == "commands":
            d = Diagnostic(
                "error", "Anonymous macro invocation requires nested commands."
            )
            return set_location(d, node)

        if isinstance(root := node.arguments[-1], AstRoot):
            namespace, resolved = self.nested_location_resolver.resolve()
            path = self.generate.format(
                f"{namespace}:{resolved}/{self.generate_macro_template}"
            )

            self.emit_function(path, root)

            command = AstCommand(
                identifier=":".join(["function", "name"] + identifier[:-1]),
                arguments=AstChildren(
                    [AstResourceLocation.from_value(path), *node.arguments[:-1]]
                ),
            )
            return set_location(command, node)

        return node

    @rule(AstCommand, identifier="schedule:function:function:time:commands")
    @rule(AstCommand, identifier="schedule:function:function:time:append:commands")
    @rule(AstCommand, identifier="schedule:function:function:time:replace:commands")
    def nesting_schedule_function(self, node: AstCommand):
        yield from self.handle_function(
            AstCommand(
                identifier="function:name:commands",
                arguments=AstChildren([node.arguments[0], node.arguments[-1]]),
            )
        )

        return replace(
            node,
            identifier=node.identifier[:-9],
            arguments=AstChildren(node.arguments[:-1]),
        )

    @rule(AstRoot)
    def nesting(self, node: AstRoot):
        changed = False
        commands: List[AstCommand] = []

        for command in node.commands:
            if command.identifier in self.identifier_map:
                result = yield from self.handle_function(command, top_level=True)
                commands.extend(result)
                changed = True
                continue

            args = command.arguments
            stack: List[AstCommand] = [command]

            expand = None

            while args and isinstance(subcommand := args[-1], AstCommand):
                if subcommand.identifier == "execute:expand:commands":
                    expand = subcommand
                    break
                stack.append(subcommand)
                args = subcommand.arguments

            if expand:
                changed = True
                for nested_command in cast(AstRoot, expand.arguments[0]).commands:
                    if nested_command.identifier == "execute:subcommand":
                        expansion = cast(AstCommand, nested_command.arguments[0])
                    else:
                        expansion = AstCommand(
                            identifier="execute:run:subcommand",
                            arguments=AstChildren([nested_command]),
                        )
                        expansion = set_location(expansion, nested_command)

                    for prefix in reversed(stack):
                        args = AstChildren([*prefix.arguments[:-1], expansion])
                        expansion = replace(prefix, arguments=args)

                    commands.append(expansion)

            else:
                commands.append(command)

        if changed:
            return replace(node, commands=AstChildren(commands))

        return node

    def handle_function(
        self,
        node: AstCommand,
        top_level: bool = False,
    ) -> Generator[Diagnostic, None, List[AstCommand]]:
        name, *args, root = node.arguments

        if isinstance(name, AstResourceLocation) and isinstance(root, AstRoot):
            path = name.get_canonical_value()

            if top_level:
                if node.identifier in (
                    "function:name:arguments:commands",
                    "function:name:with:block:sourcePos:commands",
                    "function:name:with:block:sourcePos:path:commands",
                    "function:name:with:entity:source:commands",
                    "function:name:with:entity:source:path:commands",
                    "function:name:with:storage:source:commands",
                    "function:name:with:storage:source:path:commands",
                ):
                    d = Diagnostic(
                        "error",
                        "Nested function definition can not include arguments.",
                    )
                    d.notes.append(
                        'Prefix with "execute" to invoke the nested function.'
                    )
                    yield set_location(d, node, args[-1])
                    return []
            else:
                if node.identifier == "append:function:name:commands":
                    d = Diagnostic("error", "Can't append commands with execute.")
                    yield set_location(d, node, name)
                    return []
                if node.identifier == "prepend:function:name:commands":
                    d = Diagnostic("error", "Can't prepend commands with execute.")
                    yield set_location(d, node, name)
                    return []

            target = self.database.index.get(path)

            if node.identifier in (
                "append:function:name:commands",
                "prepend:function:name:commands",
            ):
                if ":".join(self.nested_location_resolver.resolve()) == path:
                    if node.identifier == "prepend:function:name:commands":
                        d = Diagnostic(
                            "error",
                            f'Can\'t prepend commands to the current function "{path}".',
                        )
                        yield set_location(d, node, name)
                        return []
                    return list(root.commands)

                if not target:
                    self.emit_function(path, root)
                    return []

                compilation_unit = self.database[target]

                if compilation_unit.ast:
                    compilation_unit.ast = replace(
                        compilation_unit.ast,
                        commands=AstChildren(
                            compilation_unit.ast.commands + root.commands
                            if node.identifier == "append:function:name:commands"
                            else root.commands + compilation_unit.ast.commands
                        ),
                    )
                else:
                    compilation_unit.ast = root

                return []

            elif not target:
                self.emit_function(path, root)
                return []

            elif self.database[target].ast != root:
                d = Diagnostic(
                    "error",
                    f'Redefinition of function "{path}" doesn\'t match existing implementation.',
                )
                yield set_location(d, name)
                return []

            else:
                return []

        else:
            return [node]
